import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="aad-jimmielin", # Replace with your own username
    version="1.0.5",
    author="CS107-Dist-Computing-Group",
    author_email="hplin@seas.harvard.edu",
    description="Awesome Automatic Differentiation Project",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/dist-computing/cs107-FinalProject",
    packages=setuptools.find_packages(),
    install_requires=["numpy", "scipy", "pytest"],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)
